package com.emsoft.data.hibernate

import com.emsoft.common.Assert
import com.emsoft.common.StringHelper
import com.emsoft.data.SessionDataManager
import org.hibernate.*

/**
 * Manages data throught hibernate session.
 *
 * @author Vladimir Zhukov
 *
 */
class HibernateDataManager extends SessionDataManager<Session, HibernateDataManager> {

    private SessionFactory sessionFactory
    private String configFile
    protected Transaction transaction

    //    protected HibernateDataManager() {
    //    }

    HibernateDataManager(String configFile = "META-INF/hibernate.cfg.xml") {
        Assert.hasText(configFile, "configFile")
        this.configFile = configFile
    }

    HibernateDataManager(HibernateDataManager dataManager) {
        super(dataManager)
    }

    HibernateDataManager(SessionFactory sessionFactory) {
        Assert.notNull(sessionFactory, "sessionFactory")
        this.sessionFactory = sessionFactory
    }

    protected SessionFactory getFactory() {
        if (sessionFactory == null || sessionFactory.isClosed())
            sessionFactory = HibernateFactory.getFactory(configFile)
        return sessionFactory
    }

    @Override
    protected boolean rollbackTransactionCore() {
        transaction.rollback()
        return !isInTransactionCore()
    }

    @Override
    protected boolean commitTransactionCore() {
        transaction.commit()
        return !isInTransactionCore()
    }

    @Override
    protected boolean beginTransactionCore() {
        transaction = sessionObject.beginTransaction()
        return isInTransactionCore()
    }

    @Override
    protected boolean isInTransactionCore() {
        return transaction != null && transaction.isActive()
    }

    @Override
    protected boolean isOpenCore() {
        return sessionObject != null && sessionObject.open
    }

    @Override
    protected boolean openCore() {
        sessionObject = factory.openSession()
        return isOpenCore()
    }

    @Override
    protected boolean closeCore() {
        sessionObject.close()
        sessionObject = null
        return !isOpenCore()
    }

    protected <T> T attach(T obj) {
        def s = getSession()
        if (s.contains(obj)) {
            obj = s.merge(obj) as T
        } else {
            s.lock(obj, LockMode.NONE)
        }
        return obj
    }

    final protected <T> List<T> execute(Query query, Collection<T> list = null) {
        if (list == null)
            return query.list()
        def iter = query.scroll()
        try {
            while (iter.scroll(1))
                list << iter.get(0)
        }
        finally {
            iter.close()
        }
    }

    final protected Query createQuery(Class<?> entityClass, String alias, String result, String join,
            String filter, String ordering, Map<?, ?> params, int offset = 0, int limit = -1) {
        def sb = new StringBuilder()
        if (StringHelper.hasText(result))
            sb.append("select ").append(result)
        sb.append(" from ").append(entityClass.name)
        if (StringHelper.hasText(alias))
            sb.append(" as ").append(alias)
        if (StringHelper.hasText(join)) {
            sb.append(" ")
            if (!join.contains("join"))
                sb.append("join ")
            sb.append(join)
        }
        if (StringHelper.hasText(filter))
            sb.append(" where ").append(filter)
        if (StringHelper.hasText(ordering))
            sb.append(" order by ").append(ordering)
        def query = getSession().createQuery(sb.toString())
        query.firstResult = offset
        if (limit > 0)
            query.maxResults = limit
        if (params)
            setParameters(query, params)
        return query
    }

    protected void setParameters(Query query, Map<?, ?> params) {
        params.entrySet().each { item ->
            if (item.key instanceof String) {
                if (item.value && (item.value.getClass().array || item.value instanceof Collection)) {
                    query.setParameterList(item.key as String, item.value)
                } else {
                    query.setParameter(item.key as String, item.value)
                }
            } else {
                query.setParameter(item.key as int, item.value)
            }
        }
    }

    protected final int executePaging(list, queryEntities, queryTotalClosure, int offset, int limit) {
        execute(queryEntities, list)
        return offset >= 0 && limit >= 0 ? queryTotalClosure().uniqueResult() as int : list.size()
    }

    /**
     * Search for all entities of specified entityClass.
     * @param entityClass
     * @param ordering
     * @return
     */
    public <T> List<T> findAll(Class<T> entityClass, String ordering = null) {
        Assert.notNull(entityClass, "entityClass")
        execute(createQuery(entityClass, null, null, null, null, ordering, null)) as List<T>
    }

    /**
     * Search for all entities of specified entityClass.
     */
    public <T> void findAll(Collection<T> list, Class<T> entityClass, String ordering = null,
            int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.notNull(list, "list")
        execute(createQuery(entityClass, null, null, null, null, ordering, null, offset, limit), list)
    }

    /**
     * Search for all entities of specified entityClass.
     */
    public <T> int findAllPaging(Collection<T> list, Class<T> entityClass, String ordering = null,
            int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.notNull(list, "list")
        executePaging(list, createQuery(entityClass, null, null, null, null, ordering, null, offset, limit),
                { createQuery(entityClass, null, "count(*)", null, null, null, null) }, offset, limit)
    }

    /**
     * Search entities by specified filter.
     */
    public <T> List<T> find(Class<T> entityClass, String alias = null, String filter, String ordering = null,
            Map<?, ?> params = null) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(filter, "filter")
        execute(createQuery(entityClass, alias, null, null, filter, ordering, params)) as List<T>
    }

    /**
     * Search entities by specified filter.
     */
    public <T> void find(Collection<T> list, Class<T> entityClass, String alias = null, String filter,
            String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(filter, "filter")
        Assert.notNull(list, "list")
        execute(createQuery(entityClass, alias, null, null, filter, ordering, params, offset, limit), list)
    }

    /**
     * Search entities by specified filter.
     */
    public <T> int findPaging(Collection<T> list, Class<T> entityClass, String alias = null, String filter,
            String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(filter, "filter")
        Assert.notNull(list, "list")
        executePaging(list, createQuery(entityClass, alias, null, null, filter, ordering, params, offset, limit),
                { createQuery(entityClass, alias, "count(*)", null, filter, null, params) }, offset, limit)
    }

    /**
     * Search all entities and return specified result.
     */
    public <T> List<T> findAllResult(Class<?> entityClass, String result, String ordering = null,
            Map<?, ?> params = null) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        execute(createQuery(entityClass, null, result, null, null, ordering, params)) as List<T>
    }

    /**
     * Search all entities and return specified result.
     */
    void findAllResult(Collection<?> list, Class<?> entityClass, String result,
            String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        Assert.notNull(list, "list")
        execute(createQuery(entityClass, null, result, null, null, ordering, params, offset, limit), list)
    }

    /**
     * Search all entities and return specified result.
     */
    int findAllResultPaging(Collection<?> list, Class<?> entityClass, String result,
            String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        Assert.notNull(list, "list")
        executePaging(list, createQuery(entityClass, null, result, null, null, ordering, params, offset, limit),
                { createQuery(entityClass, null, "count($result)", null, null, null, params) }, offset, limit)
    }

    /**
     * Search entities by specified filter and return specified result.
     */
    public <T> List<T> findResult(Class<?> entityClass, String result, String filter,
            String ordering = null, Map<?, ?> params = null) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        Assert.hasText(filter, "filter")
        execute(createQuery(entityClass, null, result, null, filter, ordering, params)) as List<T>
    }

    /**
     * Search entities by specified filter and return specified result.
     */
    void findResult(Collection<?> list, Class<?> entityClass, String alias = null, String result,
            String filter, String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        Assert.hasText(filter, "filter")
        Assert.notNull(list, "list")
        execute(createQuery(entityClass, alias, result, null, filter, ordering, params, offset, limit), list)
    }

    /**
     * Search entities by specified filter and return specified result.
     */
    int findResultPaging(Collection<?> list, Class<?> entityClass, String alias = null, String result,
            String filter, String ordering = null, Map<?, ?> params = null, int offset = 0, int limit = -1) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        Assert.hasText(filter, "filter")
        Assert.notNull(list, "list")
        executePaging(list, createQuery(entityClass, alias, result, null, filter, ordering, params, offset, limit),
                { createQuery(entityClass, alias, "count($result)", null, filter, null, params) }, offset, limit)
    }

    /**
     * Search unique entity by specified filter.
     */
    public <T> T findUnique(Class<T> entityClass, String alias = null, String filter, Map<?, ?> params = null) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(filter, "filter")
        createQuery(entityClass, alias, null, null, filter, null, params).uniqueResult() as T
    }

    /**
     * Search unique entity by specified filter and return specified result.
     */
    public <T> T findUniqueResult(Class<?> entityClass, String alias = null, String result,
            String filter = null, Map<?, ?> params = null) {
        Assert.notNull(entityClass, "entityClass")
        Assert.hasText(result, "result")
        createQuery(entityClass, alias, result, null, filter, null, params).uniqueResult() as T
    }

    /**
     * Search entity by Id.
     */
    public <T> T findById(Class<T> entityClass, Serializable id) {
        Assert.notNull(entityClass, "entityClass")
        Assert.notNull(id, "id")
        getSession().get(entityClass, id) as T
    }

    /**
     * Load entity by Id.
     */
    public <T> T loadById(Class<T> entityClass, Serializable id) {
        Assert.notNull(entityClass, "entityClass")
        Assert.notNull(id, "id")
        getSession().load(entityClass, id) as T
    }

    /**
     * Save or update an array of instances persistent.
     */
    void save(Object... pcs) {
        Assert.hasElements(pcs, "pcs")
        def ses = getSession()
        pcs.each { ses.saveOrUpdate(it) }
    }

    /**
     * Save or update an collection of instances persistent.
     */
    void save(Collection<?> pcs, boolean flush = false) {
        Assert.hasElements(pcs, "pcs")
        save(pcs.toArray())
        if (flush)
            getSession().flush()
    }

    /**
     * Delete an array of instances from data store.
     */
    void delete(Object... pcs) {
        Assert.hasElements(pcs, "pcs")
        def s = getSession()
        pcs.each { s.delete(it) }
    }

    void delete(Collection<?> pcs, boolean flush = false) {
        Assert.hasElements(pcs, "pcs")
        delete(pcs.toArray())
        if (flush)
            getSession().flush()
    }

    /**
     *  Delete all entities of specified entityClass.
     * @return Number of deleted entities.
     */
    int delete(Class<?> entityClass, String filter = null, Map<?, ?> params = null) {
        def str = "delete from ${entityClass.name}"
        if (StringHelper.hasText(filter))
            str += " where $filter"
        def query = getSession().createQuery(str)
        if (params)
            setParameters(query, params)
        query.executeUpdate()
    }

    /**
     * Re-read obj from DB.
     */
    void refresh(def obj) {
        getSession().refresh(obj)
    }
}
