package com.emsoft.data.hibernate.types

import com.emsoft.exceptions.ExHelper
import java.lang.reflect.Method
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.SQLException
import org.hibernate.HibernateException
import org.hibernate.type.TypeResolver
import org.hibernate.usertype.ParameterizedType
import org.hibernate.usertype.UserType

class GenericEnumUserType implements UserType, ParameterizedType, Serializable {

    private static final String DEFAULT_IDENTIFIER_METHOD_NAME = "getId"
    private static final String DEFAULT_VALUE_OF_METHOD_NAME = "valueOf"

    private Class<? extends Enum> enumClass
    private Class<?> identifierType
    private Method identifierMethod
    private Method valueOfMethod
    //private SingleColumnType type
    private def type
    private int[] sqlTypes

    void setParameterValues(Properties parameters) {
        String enumClassName = parameters.getProperty("enumClassName")
        try {
            enumClass = Class.forName(enumClassName).asSubclass(Enum.class)
        }
        catch (ClassNotFoundException cfne) {
            ExHelper.throwException(new HibernateException("Enum class not found", cfne))
        }
        String identifierMethodName = parameters.getProperty("identifierMethod", DEFAULT_IDENTIFIER_METHOD_NAME)
        try {
            identifierMethod = enumClass.getMethod(identifierMethodName)
        }
        catch (ex) {
            ExHelper.throwException(new HibernateException("Failed to obtain identifier method", ex))
        }
        identifierType = identifierMethod.getReturnType()

        //type = (NullableType) TypeFactory.basic(identifierType.name)
        type = new TypeResolver().heuristicType(identifierType.getName(), parameters)

        if (type == null)
            ExHelper.throwException(
                    new HibernateException("Unsupported identifier type ${identifierType.name}"))

        sqlTypes = [type.sqlType()]

        String valueOfMethodName = parameters.getProperty("valueOfMethod", DEFAULT_VALUE_OF_METHOD_NAME)

        try {
            valueOfMethod = enumClass.getMethod(valueOfMethodName, identifierType)
        }
        catch (ex) {
            ExHelper.throwException(new HibernateException("Failed to obtain valueOf method", ex))
        }
    }

    Class returnedClass() {
        return enumClass
    }

    def nullSafeGet(ResultSet rs, String[] names, def owner) throws HibernateException, SQLException {
        def identifier = type.get(rs, names[0])
        if (rs.wasNull())
            return null
        try {
            return valueOfMethod.invoke(enumClass, identifier)
        }
        catch (ex) {
            ExHelper.throwException(new HibernateException(
                    "Exception while invoking valueOf method '${valueOfMethod.getName()}' of " +
                            "enumeration class '${enumClass}'", ex))
        }
    }

    void nullSafeSet(PreparedStatement st, def value, int index) throws HibernateException, SQLException {
        try {
            if (value == null) {
                st.setNull(index, type.sqlType())
            }
            else {
                def identifier = identifierMethod.invoke(value)
                type.set(st, identifier, index)
            }
        }
        catch (ex) {
            ExHelper.throwException(new HibernateException(
                    "Exception while invoking identifierMethod '${identifierMethod.getName()}' of " +
                            "enumeration class '${enumClass}'", ex))
        }
    }

    int[] sqlTypes() {
        return sqlTypes
    }

    def assemble(Serializable cached, def owner) throws HibernateException {
        return cached
    }

    def deepCopy(def value) throws HibernateException {
        return value
    }

    Serializable disassemble(def value) throws HibernateException {
        return value
    }

    boolean equals(def x, def y) throws HibernateException {
        return x == y
    }

    int hashCode(def x) throws HibernateException {
        return x.hashCode()
    }

    boolean isMutable() {
        return false
    }

    def replace(def original, def target, def owner) throws HibernateException {
        return original
    }
}